﻿using DetourFunc;
using DetourFunc.Clr;
using System;
using System.Runtime.InteropServices;

namespace ConsoleApp1
{
    [UnmanagedFunctionPointer(CallingConvention.StdCall)]
    public delegate void SleepExDelegate(int milliseconds, bool bAlertable);

    class Program
    {
        [DllImport("kernel32", SetLastError = true, CharSet = CharSet.Ansi, BestFitMapping = false)]
        internal static extern IntPtr LoadLibrary(string lpFileName);

        [DllImport("kernel32", CharSet = CharSet.Ansi, ExactSpelling = true, SetLastError = true, BestFitMapping = false)]
        internal static extern IntPtr GetProcAddress(IntPtr hModule, string procName);

        [DllImport("kernel32.dll", CallingConvention = CallingConvention.StdCall)]
        public static extern void SleepEx(int milliseconds, bool bAlertable);

        static SleepExDelegate s_originalSleepExFunc;

        static void Main(string[] _)
        {
            IntPtr sleepPtr = MethodReplacer.GetExportFunctionAddress("kernel32.dll", "SleepEx", out var _);

            //IntPtr modulePtr = LoadLibrary("kernel32.dll");
            //if (modulePtr != IntPtr.Zero)
            //{
            //    IntPtr funcAddr = GetProcAddress(modulePtr, "SleepEx");
            //    Console.WriteLine($"SleepEx: {funcAddr.ToInt64():x}");
            //}

            IntPtr ptrBodyReplaceMethod;

            {
                SleepExDelegate action = Replaced_TestMethod;
                MethodDesc mdReplaceMethod = MethodDesc.ReadFromMethodInfo(action.Method);
                ptrBodyReplaceMethod = mdReplaceMethod.GetNativeFunctionPointer();
            }

            Console.WriteLine($"Address to be patched: {sleepPtr.ToInt64():x}");
            Console.WriteLine($"With this address: {ptrBodyReplaceMethod.ToInt64():x}");

            using (var item = new TrampolinePatch<SleepExDelegate>())
            {
                if (item.JumpPatch(sleepPtr, ptrBodyReplaceMethod) == true)
                {
                    s_originalSleepExFunc = item.GetOriginalFunc();
                }

                SleepEx(3000, false);
            }

            Console.WriteLine("Press any key to exit...");
            Console.ReadLine();
        }

        public static void Replaced_TestMethod(int milliseconds, bool bAlertable)
        {
            Console.WriteLine($"Replaced_TestMethod called: milliseconds = {milliseconds}, bAlertable = {bAlertable}");
            s_originalSleepExFunc?.Invoke(3000, false);
        }
    }
}
